# =============================================================================
#
# Copyright (c)  2016, Cisco Systems
# All rights reserved.
#
# # Author: Klaudiusz Staniek
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
# Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
# THE POSSIBILITY OF SUCH DAMAGE.
# =============================================================================

import ConfigParser
import getpass
import fnmatch

try:
    import keyring
except ImportError:
    pass


def make_realm(name):
    return "Condoor@{}".format(name)


class AccountManager(object):
    def __init__(self,
                 config_file='accounts.cfg',
                 username_cb=None,
                 password_cb=None):
        self.config_file = config_file
        self.config = ConfigParser.SafeConfigParser({'username': '', })
        self.config.read(self.config_file)
        self.config.write(open(self.config_file, 'w'))

        self.username_cb = username_cb \
            if callable(username_cb) else self._prompt_for_username
        self.password_cb = password_cb \
            if callable(password_cb) else self._prompt_for_password

    def _prompt_for_username(self, prompt):
        # Not sure needed
        return None

    def _prompt_for_password(self, prompt):
        return getpass.getpass(prompt)

    def _find_section(self, realm):
        for section in iter(self.config.sections()):
            if fnmatch.fnmatch(realm, section):
                break
        else:
            section = 'DEFAULT'
        return section

    def _get_username(self, section):
        username = self.config.get(section, 'username')
        if username == '':
            return None
        return username

    def get_username(self, realm):
        username = self._get_username(self._find_section(realm))
        if not username:
            username = getpass.getuser()
        return username

    def get_password(self, realm, username=None, interact=True):
        section = self._find_section(realm)
        config_user_name = self._get_username(section)
        if not config_user_name or username != config_user_name:
                return None

        if not username:
            username = config_user_name

        try:
            password = keyring.get_password(make_realm(section), username)
        except:
            password = None

        if password is None and interact:
            prompt = "{}@{} Password: ".format(username, realm)
            password = self.password_cb(prompt)
            self.set_password(make_realm(section), username, password)
        return password

    def set_password(self, realm, username, password):
        try:
            keyring.set_password(realm, username, password)
        except:
            pass

    def get_login(self, realm):
        username = self.get_username(realm)
        password = self.get_password(realm, username)
        return username, password



# =============================================================================
# controllers
#
# Copyright (c)  2016, Cisco Systems
# All rights reserved.
#
# # Author: Klaudiusz Staniek
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
# Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
# THE POSSIBILITY OF SUCH DAMAGE.
# =============================================================================

import logging

from time import sleep

from ..utils import to_list
from ..utils import delegate

from ..controllers.protocols import make_protocol
from ..exceptions import ConnectionError

import pexpect


# Delegate following methods to _session class
@delegate("_session", ("expect", "expect_exact", "sendline",
                       "isalive", "sendcontrol", "send", "read_nonblocking", "setecho"))
class Controller(object):
    def __init__(self, platform, hostname, hosts, account_manager=None, max_attempts=1, logfile=None):
        self.hosts = to_list(hosts)
        self.max_attempts = max_attempts
        self.account_mgr = account_manager
        self.session_log = logfile
        self.hostname = hostname
        self.platform = platform
        self.connected = False
        self.authenticated = False
        self._session = None
        self.detected_prompts = []
        self.is_target = False
        self.last_hop = 0
        self.last_pattern = None
        self.logger = logging.getLogger("condoor.controller")
        self._clear_detected_prompts()

    @property
    def before(self):
        """
        Property added to imitate pexpect.spawn class
        """
        return self._session.before if self._session else None

    @property
    def after(self):
        """
        Property added to imitate pexpect.spawn class
        """
        return self._session.after if self._session else None

    @property
    def detected_target_prompt(self):
        """
        Detected target device prompt detected by controller
        """
        return self.detected_prompts[-1] if len(self.hosts) > 0 else ""

    @detected_target_prompt.setter
    def detected_target_prompt(self, prompt):
        target_hop = len(self.hosts)
        self._dbg(10, "[{}] {}: Updated target prompt: {}".format(target_hop, self.hosts[-1].hostname, prompt))
        self.detected_prompts[-1] = prompt

    def connect(self, start_hop=0, spawn=True, detect_prompt=True):
        # it can restart from the last hop.

        connected = self.connected
        if connected:
            spawn = False

        if self.last_hop > 0 and start_hop == 0:
            start_hop = self.last_hop

        hosts = self.hosts[self.last_hop:]
        host_count = len(hosts)

        self._dbg(10, "Restarting from hop: {}".format(start_hop))

        for hop, host in enumerate(hosts, start=start_hop+1):
            if hop == host_count:
                self.is_target = True

            if not host.is_valid():
                raise ConnectionError("Invalid host", host)
            attempt = 1
            while attempt <= self.max_attempts:
                if not host.is_reachable():
                    self._dbg(40, "[{}] {}: Host not reachable".format(hop, host.hostname))
                else:
                    self._dbg(
                        10,
                        "[{}] {}: Connecting. Attempt ({}/{})".format(hop, host.hostname, attempt, self.max_attempts)
                    )
                    try:
                        if self.is_target:
                            self._dbg(10, "[{}] {}: Connecting to target device".format(hop, host.hostname))
                        else:
                            self._dbg(10, "[{}] {}: Connecting to jump host".format(hop, host.hostname))

                        protocol = make_protocol(self, host, spawn, self.account_mgr, self.session_log)
                        if protocol.connect():
                            if protocol.authenticate(self.detected_prompts[hop]):
                                connected = True
                                if detect_prompt:
                                    if not protocol.detect_prompt():
                                        connected = False
                        else:
                            connected = False
                    except Exception as e:
                        self._dbg(40, "Error during connecting to device: {}".format(e.message))
                        raise

                    if connected:
                        self.detected_prompts[hop] = protocol.prompt
                        break

                attempt += 1
                sleep(2)
            else:
                self._dbg(40, "[{}] {}: Connection error. ""Max attempts reached.".format(hop, host.hostname))
                self.disconnect()
                raise ConnectionError(host=self.hostname)

            self._dbg(10, "[{}] {}: Connected successfully".format(hop, host.hostname))

        if connected:
            self._dbg(10, "Connected target device")
            self.connected = True

        return connected

    def disconnect(self):
        """
        Gracefully disconnect from all the nodes
        """
        self._dbg(10, "Initializing the disconnection process")
        if self._session and self.isalive():
            self._dbg(10, "Disconnecting the sessions")
            index = 0
            hop = 0
            while index != 1 and hop < 10:
                self.sendline('exit')
                index = self.expect(
                    [pexpect.TIMEOUT, pexpect.EOF, "con.*is now available|rommon"],
                    timeout=2
                )
                if index == 1:
                    break

                if index == 2:  # console connected through TS
                    self._dbg(10, "Console connection detected")
                    self.sendline('\x03')
                    self.sendcontrol(']')
                    self.sendline('quit')

                hop += 1

        self._session.close()
        self._dbg(10, "Disconnected")
        self.last_hop = 0
        self.last_pattern = None
        self.connected = False

    def _dbg(self, level, msg):
        self.logger.log(level, "[{}]: [CTRL] {}".format(self.hostname, msg))

    def _clear_detected_prompts(self):
        self.detected_prompts = []
        for i in xrange(len(self.hosts) + 1):
            self.detected_prompts.append(None)
        self.detected_prompts[0] = "FaKePrOmPt"


# =============================================================================
#
# Copyright (c)  2016, Cisco Systems
# All rights reserved.
#
# # Author: Klaudiusz Staniek
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
# Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
# THE POSSIBILITY OF SUCH DAMAGE.
# =============================================================================

import socket
import time


def delegate(attribute_name, method_names):
    """Passes the call to the attribute called attribute_name for
    every method listed in method_names.
    """
    # hack for python 2.7 as nonlocal is not available
    d = {
        'attribute': attribute_name,
        'methods': method_names
    }

    def decorator(cls):
        attribute = d['attribute']
        if attribute.startswith("__"):
            attribute = "_" + cls.__name__ + attribute
        for name in d['methods']:
            setattr(cls, name, eval("lambda self, *a, **kw: "
                                    "self.{0}.{1}(*a, **kw)".format(attribute, name)))
        return cls
    return decorator


def to_list(item):
    """
    If the given item is iterable, this function returns the given item.
    If the item is not iterable, this function returns a list with only the
    item in it.

    @type  item: object
    @param item: Any object.
    @rtype:  list
    @return: A list with the item in it.
    """
    if hasattr(item, '__iter__'):
        return item
    return [item]


def is_reachable(host, port=23):
    """
    This function check reachability for specified hostname/port
    It tries to open TCP socket.
    It supports IPv6.
    :param host: hostname or ip address string
    :rtype: str
    :param port: tcp port number
    :rtype: number
    :return: True if host is reachable else false
    """

    try:
        addresses = socket.getaddrinfo(
            host, port, socket.AF_UNSPEC, socket.SOCK_STREAM
        )
    except socket.gaierror:
        return False

    for family, socktype, proto, cannonname, sockaddr in addresses:
        sock = socket.socket(family, socket.SOCK_STREAM)
        sock.settimeout(5)
        try:
            sock.connect(sockaddr)
        except IOError:
            continue

        sock.shutdown(socket.SHUT_RDWR)
        sock.close()
        # Wait 2 sec for socket to shutdown
        time.sleep(2)
        break
    else:
        return False
    return True



# =============================================================================
# protocol
#
# Copyright (c)  2016, Cisco Systems
# All rights reserved.
#
# # Author: Klaudiusz Staniek
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
# Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
# THE POSSIBILITY OF SUCH DAMAGE.
# =============================================================================

import re
import pexpect
import time
import logging

from ...exceptions import \
    ConnectionAuthenticationError, \
    ConnectionError, \
    ConnectionTimeoutError

from ..fsm import action


# used for unix jumphosts
PASSWORD_PROMPT = re.compile("[P|p]assword:\s?")
USERNAME_PROMPT = re.compile("([U|u]sername:|login:)\s?")
SHELL_PROMPT = re.compile("\$\s?|>\s?|#\s?|\%\s?")

PERMISSION_DENIED = "Permission denied"
AUTH_FAILED = "Authentication failed|not authorized|Login incorrect"
RESET_BY_PEER = "reset by peer|closed by foreign host"
RECONFIGURE_USERNAME_PROMPT = "[Nn][Oo] root-system username is configured"
SET_USERNAME = "[Ee]nter.*username:"
SET_PASSWORD = "Enter secret"
PASSWORD_OK = "[Pp]assword [Oo][Kk]"
PRESS_RETURN = "Press RETURN to get started\."
STANDBY_CONSOLE = "Standby console disabled"

# Error when the hostname can't be resolved or there is
# network reachability timeout
UNABLE_TO_CONNECT = "nodename nor servname provided, or not known|" \
                    "Unknown host|" \
                    "[Operation|Connection] timed out|" \
                    "[D|d]estination unreachable|" \
                    "[U|u]nable to connect|" \
                    "[C|c]onnection refused"


class Protocol(object):

    def __init__(self, controller, node_info, account_manager=None, logfile=None):
        self.protocol = node_info.protocol
        self.hostname = node_info.hostname
        self.port = node_info.port
        self.password = node_info.password

        self.ctrl = controller
        self.logfile = logfile
        self.account_manager = account_manager

        username = node_info.username
        if not username and self.account_manager:
            username = self.account_manager.get_username(self.hostname)

        self.username = username
        self.prompt = ""
        self.last_pattern = None
        self.logger = logging.getLogger("condoor.controller.protocol")

    def _spawn_session(self, command):
        self._dbg(10, "Executing command: '{}'".format(command))
        if self.ctrl._session and self.ctrl.isalive():
            try:
                self.ctrl.send(command)
                self.ctrl.expect_exact(command, timeout=20)
                self.ctrl.sendline()

            except pexpect.EOF:
                raise ConnectionError("Connection error", self.hostname)
            except pexpect.TIMEOUT:
                raise ConnectionTimeoutError("Timeout", self.hostname)

        else:
            try:
                self.ctrl._session = pexpect.spawn(
                    command,
                    maxread=50000,
                    searchwindowsize=None,
                    echo=True  # KEEP YOUR DIRTY HANDS OFF FROM ECHO!
                )
            except pexpect.EOF:
                raise ConnectionError("Connection error", self.hostname)
            except pexpect.TIMEOUT:
                raise ConnectionTimeoutError("Timeout", self.hostname)

            self.ctrl._session.logfile_read = self.logfile

    def connect(self):
        """
        Protocol specific implementation
        """
        raise NotImplementedError("Connection method not implemented")

    def authenticate(self):
        """
        Protocol specific implementation
        """
        raise NotImplementedError("Authentication method not implemented")

    def disconnect(self):
        """
        Protocol specific implementation
        """
        raise NotImplementedError("Authentication method not implemented")

    def try_read_prompt(self, timeout_multiplier):
        """
        based on try_read_prompt from pxssh.py
        https://github.com/pexpect/pexpect/blob/master/pexpect/pxssh.py
        """
        # maximum time allowed to read the first response
        first_char_timeout = timeout_multiplier * 2

        # maximum time allowed between subsequent characters
        inter_char_timeout = timeout_multiplier * 0.4

        # maximum time for reading the entire prompt
        total_timeout = timeout_multiplier * 4

        prompt = ""
        begin = time.time()
        expired = 0.0
        timeout = first_char_timeout

        while expired < total_timeout:
            try:
                p = self.ctrl.read_nonblocking(size=1, timeout=timeout)
                # \r=0x0d CR \n=0x0a LF
                if p not in ['\n', '\r']:  # omit the cr/lf sent to get the prompt
                    timeout = inter_char_timeout
                expired = time.time() - begin
                prompt += p
            except pexpect.TIMEOUT:
                break
            except pexpect.EOF:
                raise ConnectionError('Session disconnected')

        #  print("expired:{},total:{}".format(expired, total_timeout))
        #  print(":".join("{:02x}".format(ord(c)) for c in prompt))
        #  print("RAW: {}".format(repr(prompt)))
        #  print prompt.split('\r')
        #  print prompt.strip()
        #  prompt = prompt.split('\r')[-1].strip()
        prompt = prompt.strip()
        #  print("PROMPT: '%s'" % prompt)
        return prompt

    def levenshtein_distance(self, a, b):
        """
        This calculates the Levenshtein distance between string a and b.

        :param a: String - input string a
        :param b: String - input string b
        :return: Number - Levenshtein Distance between string a and b
        """

        n, m = len(a), len(b)
        if n > m:
            a, b = b, a
            n, m = m, n
        current = range(n+1)
        for i in range(1, m+1):
            previous, current = current, [i]+[0]*n
            for j in range(1, n+1):
                add, delete = previous[j]+1, current[j-1]+1
                change = previous[j-1]
                if a[j-1] != b[i-1]:
                    change += + 1
                current[j] = min(add, delete, change)
        return current[n]

    def detect_prompt(self, sync_multiplier=4):
        """
        This attempts to find the prompt. Basically, press enter and record
        the response; press enter again and record the response; if the two
        responses are similar then assume we are at the original prompt.
        This can be a slow function. Worst case with the default sync_multiplier
        can take 16 seconds. Low latency connections are more likely to fail
        with a low sync_multiplier. Best case sync time gets worse with a
        high sync multiplier (500 ms with default).

        """
        self.ctrl.sendline()
        self.try_read_prompt(sync_multiplier)

        attempt = 0
        max_attempts = 10
        while attempt < max_attempts:
            attempt += 1
            self._dbg(10, "Detecting prompt. Attempt ({}/{})".format(attempt, max_attempts))

            self.ctrl.sendline()
            a = self.try_read_prompt(sync_multiplier)

            self.ctrl.sendline()
            b = self.try_read_prompt(sync_multiplier)

            ld = self.levenshtein_distance(a, b)
            len_a = len(a)
            self._dbg(10, "LD={},MP={}".format(ld, sync_multiplier))
            sync_multiplier *= 1.2
            if len_a == 0:
                continue

            if float(ld)/len_a < 0.3:
                self.prompt = b.splitlines(True)[-1]
                self._dbg(10, "Detected prompt: '{}'".format(self.prompt))
                compiled_prompt = re.compile("(\r\n|\n\r){}".format(re.escape(self.prompt)))
                self._dbg(10, "Compiled prompt: '{}".format(repr(compiled_prompt.pattern)))
                self.ctrl.sendline()
                self.ctrl.expect(compiled_prompt)  # match from new line
                return True

        return False

    def _acquire_password(self):
        password = self.password
        if not password:
            if self.account_manager:
                self._dbg(
                    20,
                    "{}: {}: Acquiring password for {} "
                    "from system KeyRing".format(
                        self.protocol, self.hostname, self.username)
                )
                password = self.account_manager.get_password(
                    self.hostname,
                    self.username,
                    interact=True
                )
                if not password:
                    self._dbg(
                        30,
                        "{}: {}: Password for {} does not exists "
                        "in KeyRing".format(
                            self.protocol, self.hostname, self.username)
                    )
        return password

    @action
    def send_username(self, ctx):
        ctx.ctrl.sendline(self.username)
        ctx.timeout = 10
        return True

    @action
    def send_pass(self, ctx):
        password = self._acquire_password()
        if password:
            ctx.ctrl.setecho(False)
            ctx.ctrl.sendline(password)
            ctx.ctrl.setecho(True)
            ctx.timeout = 30
            return True
        else:
            self.disconnect()
            raise ConnectionAuthenticationError("Password not provided", self.hostname)

    @action
    def try_detect_prompt(self, ctx):
        ctx.finished = True
        if self.detect_prompt():
            return True
        else:
            return False

    @action
    def authentication_error(self, ctx):
        self.disconnect()
        raise ConnectionAuthenticationError("Authentication failed", self.hostname)

    @action
    def unable_to_connect(self, ctx):
        ctx.msg = "{}{}".format(self.ctrl.before, self.ctrl.after)
        return False

    @action
    def session_closed(self, ctx):
        ctx.msg = "First session closed"
        ctx.failed = True
        return False

    @action
    def send_new_line(self, ctx):
        # print(":".join("{:02x}".format(ord(c)) for c in ctx.ctrl.before))
        ctx.ctrl.send("\r\n")
        return True

    @action
    def save_pattern(self, ctx):
        self.last_pattern = ctx.pattern
        return True

    @action
    def send_q(self, ctx):
        ctx.ctrl.send("q")
        return True


# =============================================================================
# telnet
#
# Copyright (c)  2016, Cisco Systems
# All rights reserved.
#
# # Author: Klaudiusz Staniek
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
# Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
# THE POSSIBILITY OF SUCH DAMAGE.
# =============================================================================

from base import *
from ..fsm import FSM, action

from ...exceptions import \
    ConnectionError, \
    ConnectionTimeoutError

# Telnet connection initiated
ESCAPE_CHAR = "Escape character is|Open"
# Connection refused i.e. line busy on TS
CONNECTION_REFUSED = re.compile("Connection refused")


class Telnet(Protocol):
    def __init__(self, controller, device, spawn, account_manager, logfile):
        super(Telnet, self).__init__(controller,
                                     device,
                                     account_manager,
                                     logfile)
        command = "telnet {} {}".format(
            self.hostname, self.port
        )
        if spawn:
            self._spawn_session(command)

    def connect(self):

        if self.ctrl.is_target:
            prompt = self.ctrl.platform.platform_prompt
            rommon_prompt = self.ctrl.platform.rommon_prompt
            more = self.ctrl.platform.more
            password_prompt = self.ctrl.platform.password_prompt
            username_prompt = self.ctrl.platform.username_prompt

        else:
            prompt = SHELL_PROMPT
            rommon_prompt = SHELL_PROMPT
            more = "!@#!@#"  # find another solution
            password_prompt = PASSWORD_PROMPT
            username_prompt = USERNAME_PROMPT
        #              0            1               2                3              4            5      6
        events = [ESCAPE_CHAR, PRESS_RETURN, STANDBY_CONSOLE, username_prompt, password_prompt, more, prompt,
                  #     7                8                 9               10
                  rommon_prompt, UNABLE_TO_CONNECT, RESET_BY_PEER, pexpect.TIMEOUT]

        transitions = [
            (ESCAPE_CHAR, [0], 1, None, 20),
            (PRESS_RETURN, [0, 1], 1, self.send_new_line, 10),
            (STANDBY_CONSOLE, [0, 5], -1, ConnectionError("Standby console", self.hostname), 0),
            (username_prompt, [0, 1, 5, 6], -1, self.save_pattern, 0),
            (password_prompt, [0, 1, 5], -1, self.save_pattern, 0),
            (more, [0, 5], 7, self.send_q, 10),
            # router sends it again to delete
            (more, [7], 8, None, 10),
            # (prompt, [0, 1, 5], 6, self.send_new_line, 10),
            (prompt, [0, 1, 5], 0, None, 10),
            (prompt, [6, 8, 5], -1, self.save_pattern, 0),
            (rommon_prompt, [0, 1], -1, self.save_pattern, 0),
            (UNABLE_TO_CONNECT, [0], -1, self.unable_to_connect, 0),
            (RESET_BY_PEER, [0, 1], -1, self.unable_to_connect, 0),
            (pexpect.TIMEOUT, [0, 1], 5, self.send_new_line, 10),
            (pexpect.TIMEOUT, [5], -1, ConnectionTimeoutError("Connection timeout", self.hostname), 0)
        ]
        sm = FSM("TELNET-CONNECT", self.ctrl, events, transitions, init_pattern=self.ctrl.last_pattern)
        return sm.run()

    def authenticate(self, prompt=None):
        if self.ctrl.is_target:
            prompt = self.ctrl.platform.platform_prompt
            rommon_prompt = self.ctrl.platform.rommon_prompt
            password_prompt = self.ctrl.platform.password_prompt
            username_prompt = self.ctrl.platform.username_prompt
        else:
            if prompt is None:
                prompt = SHELL_PROMPT
            rommon_prompt = SHELL_PROMPT
            password_prompt = PASSWORD_PROMPT
            username_prompt = USERNAME_PROMPT

        events = [username_prompt, password_prompt, prompt, rommon_prompt, UNABLE_TO_CONNECT, CONNECTION_REFUSED,
                  RESET_BY_PEER, PERMISSION_DENIED, AUTH_FAILED,
                  pexpect.TIMEOUT, pexpect.EOF]

        transitions = [
            (username_prompt, [0, 1], 1, self.send_username, 10),
            (password_prompt, [0, 1], 2, self.send_pass, 20),
            (username_prompt, [2], -1, self.authentication_error, 0),
            (prompt, [0, 1, 2], -1, None, 0),
            (rommon_prompt, [0], -1, self.send_new_line, 0),
            (pexpect.TIMEOUT, [0], 1, self.send_new_line, 10),
            (pexpect.TIMEOUT, [2], -1, None, 0),
            (pexpect.TIMEOUT, [6], -1, ConnectionError("Unable to get shell prompt", self.hostname), 0),
            (pexpect.TIMEOUT, [3, 7], -1,  ConnectionTimeoutError("Connection Timeout", self.hostname), 0),
        ]

        if isinstance(prompt, str):
            self._dbg(10, "EXPECTED_PROMPT={}".format(prompt))
        elif prompt is not None:
            self._dbg(10, "EXPECTED_PROMPT={}".format(prompt.pattern))

        sm = FSM("TELNET-AUTH", self.ctrl, events, transitions, init_pattern=self.last_pattern)
        self.try_read_prompt(1)
        return sm.run()

    @action
    def set_custom_prompt_ps1(self, ctx):
        ctx.ctrl.sendline('PS1="{}"'.format("CONDORPROMPT"))
        ctx.timeout = 5
        return True

    @action
    def set_custom_prompt(self, ctx):
        ctx.ctrl.sendline('set prompt="{}"'.format("CONDORPROMPT"))
        ctx.timeout = 5
        return True

    @action
    def error(self, ctx):
        ctx.failed = True
        return False

    def disconnect(self):
        self.ctrl.sendcontrol(']')
        self.ctrl.sendline('quit')

    def _dbg(self, level, msg):
        self.logger.log(
            level, "[{}]: [TELNET]: {}".format(self.ctrl.hostname, msg)
        )